!-----------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations         !
!   Copyright (C) 2000 - 2009  CP2K developers group                          !
!-----------------------------------------------------------------------------!

! *****************************************************************************
!> \brief Types and set/get functions for HFX
!> \par History
!>      04.2008 created [Manuel Guidon]
!> \author Manuel Guidon 
! *****************************************************************************
MODULE hfx_types
  USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                             get_atomic_kind,&
                                             get_atomic_kind_set
  USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                             gto_basis_set_type
  USE bibliography,                    ONLY: cite_reference,&
                                             guidon2008
  USE cell_types,                      ONLY: cell_type,&
                                             get_cell,&
                                             plane_distance,&
                                             scaled_to_real
  USE cp_control_types,                ONLY: dft_control_type
  USE cp_files,                        ONLY: close_file,&
                                             open_file
  USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                             cp_print_key_unit_nr
  USE cp_para_types,                   ONLY: cp_para_env_type
  USE cp_units,                        ONLY: cp_unit_from_cp2k
  USE f77_blas
  USE hfx_helpers,                     ONLY: count_cells_perd,&
                                             next_image_cell_perd
  USE hfx_libint_wrapper,              ONLY: lib_int
  USE input_constants,                 ONLY: &
       do_hfx_auto_shells, do_hfx_potential_coulomb, &
       do_hfx_potential_gaussian, do_hfx_potential_id, do_hfx_potential_long, &
       do_hfx_potential_mix_cl, do_hfx_potential_mix_lg, &
       do_hfx_potential_short, do_hfx_potential_truncated, &
       use_aux_fit_basis_set
  USE input_section_types,             ONLY: section_vals_get,&
                                             section_vals_get_subs_vals,&
                                             section_vals_type,&
                                             section_vals_val_get
  USE kinds,                           ONLY: default_path_length,&
                                             dp,&
                                             int_8
  USE machine,                         ONLY: m_chdir,&
                                             m_getcwd
  USE string_utilities,                ONLY: compress
  USE t_c_g0,                          ONLY: free
  USE timings,                         ONLY: timeset,&
                                             timestop
#include "cp_common_uses.h"

  IMPLICIT NONE
  PRIVATE
  PUBLIC  hfx_type, hfx_create, hfx_release,& 
          hfx_set_distr_energy, &
          hfx_set_distr_forces, &
          hfx_cell_type, hfx_distribution, &
          hfx_potential_type, hfx_screening_type, hfx_periodic_type,&
          hfx_memory_type, hfx_load_balance_type, hfx_general_type,&
          hfx_compression_type,&
          hfx_container_type, hfx_container_node, hfx_cache_type, &
          hfx_basis_type, parse_memory_section, &
          hfx_p_type, hfx_init_container,&
          hfx_basis_info_type, hfx_screen_coeff_type,&
          hfx_reset_memory_usage_counter, pair_list_type, pair_list_element_type, pair_set_list_type

#define CACHE_SIZE 1024

  CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'hfx_types'
  INTEGER(int_8), PARAMETER, PRIVATE         :: one = 1_int_8 
  INTEGER, PARAMETER, PUBLIC                 :: max_atom_block=32
  REAL(dp), PARAMETER, PUBLIC                :: log_zero = -1000.0_dp
  REAL(dp), PARAMETER, PUBLIC                :: powell_min_log = -20.0_dp
  REAL(KIND=dp), DIMENSION(0:10), &
     PARAMETER, PUBLIC                       :: mul_fact = (/1.0_dp,&
                                                             1.1781_dp,&
                                                             1.3333_dp,&
                                                             1.4726_dp,&
                                                             1.6000_dp,&
                                                             1.7181_dp,&
                                                             1.8286_dp,&
                                                             1.9328_dp,&
                                                             2.0317_dp,&
                                                             2.1261_dp,&
                                                             2.2165_dp/)
!***

! *****************************************************************************
  TYPE hfx_potential_type
    INTEGER                                  :: potential_type     !! 1/r/ erfc(wr)/r ...
    REAL(dp)                                 :: omega              !! w 
    REAL(dp)                                 :: scale_coulomb      !! scaling factor for mixed potential
    REAL(dp)                                 :: scale_longrange    !! scaling factor for mixed potential
    REAL(dp)                                 :: scale_gaussian     !! scaling factor for mixed potential
    REAL(dp)                                 :: cutoff_radius      !! cutoff radius if cutoff potential in use
    CHARACTER(default_path_length)           :: filename
  END TYPE

! *****************************************************************************
  TYPE hfx_screening_type
    REAL(dp)                                 :: eps_schwarz         !! threshold
    REAL(dp)                                 :: eps_schwarz_forces  !! threshold
    LOGICAL                                  :: do_p_screening_forces  !! screen on P^2 ?
    LOGICAL                                  :: do_initial_p_screening !! screen on inital guess?
  END TYPE
  
! *****************************************************************************
  TYPE hfx_memory_type
    INTEGER                                  :: max_memory              !! user def max memory MB
    INTEGER(int_8)                           :: max_compression_counter !! corresponding number of reals 
    LOGICAL                                  :: do_all_on_the_fly       !! max mem == 0 ?
    REAL(dp)                                 :: eps_storage_scaling
    INTEGER                                  :: cache_size
    INTEGER                                  :: actual_memory_usage
    INTEGER                                  :: actual_memory_usage_disk
    INTEGER(int_8)                           :: max_compression_counter_disk
    LOGICAL                                  :: do_disk_storage
    CHARACTER(len=default_path_length)       :: storage_location
    INTEGER(int_8)                           :: ram_counter
  END TYPE

! *****************************************************************************
  TYPE hfx_periodic_type
    INTEGER                                  :: number_of_shells        !! number of periodic image cells 
    LOGICAL                                  :: do_periodic             !! periodic ?
    INTEGER                                  :: perd(3)                 !! x,xy,xyz,...
    INTEGER                                  :: mode
  END TYPE

! *****************************************************************************
  TYPE hfx_load_balance_type
    INTEGER                                  :: nbins                   
    INTEGER                                  :: block_size
  END TYPE

! *****************************************************************************
  TYPE hfx_general_type
    REAL(dp)                                 :: fraction                 !! for hybrids
    LOGICAL                                  :: treat_lsd_in_core
  END TYPE

! *****************************************************************************
  TYPE hfx_cell_type
    REAL(dp)                                 :: cell(3)
    REAL(dp)                                 :: cell_r(3)
  END TYPE

! *****************************************************************************
  TYPE hfx_distribution
    INTEGER(int_8)                           :: istart, jstart, kstart, lstart
    INTEGER(int_8)                           :: number_of_atom_quartets
    INTEGER(int_8)                           :: cost
    INTEGER(int_8)                           :: original_cpu_id
    REAL(KIND=dp)                            :: time_first_scf
    REAL(KIND=dp)                            :: time_other_scf
    REAL(KIND=dp)                            :: time_forces
  END TYPE

! *****************************************************************************
  TYPE pair_list_element_type
      INTEGER, DIMENSION(2) :: pair
      INTEGER, DIMENSION(2) :: set_bounds
      INTEGER, DIMENSION(2) :: kind_pair
      REAL(KIND=dp)         :: r1(3),r2(3)
      REAL(KIND=dp)         :: vec(3)
      REAL(KIND=dp)         :: dist2
  END TYPE

  ! *****************************************************************************
  TYPE pair_set_list_type
      INTEGER, DIMENSION(2) :: pair
  END TYPE

! *****************************************************************************
  TYPE pair_list_type
    TYPE(pair_list_element_type), DIMENSION(max_atom_block**2) :: elements
    INTEGER :: n_element
  END TYPE pair_list_type
  
! *****************************************************************************
  TYPE hfx_compression_type
    INTEGER(int_8), DIMENSION(:), POINTER    :: compressed_data
    INTEGER(int_8)                           :: element_pointer
    INTEGER                                  :: bit_pointer
  END TYPE
 
! *****************************************************************************
  TYPE hfx_cache_type
    INTEGER(int_8), DIMENSION(CACHE_SIZE)    :: DATA
    INTEGER                                  :: element_counter
  END TYPE

! *****************************************************************************
  TYPE hfx_container_node
    TYPE(hfx_container_node), POINTER        :: next, prev
    INTEGER(int_8), DIMENSION(CACHE_SIZE)    :: DATA
  END TYPE

! *****************************************************************************
  TYPE hfx_container_type
    TYPE(hfx_container_node), POINTER        :: first, current    
    INTEGER                                  :: element_counter
    INTEGER(int_8)                           :: file_counter
    CHARACTER(LEN=5)                         :: desc
    INTEGER                                  :: unit
    CHARACTER(default_path_length)           :: filename
  END TYPE

! *****************************************************************************
  TYPE hfx_basis_type
    INTEGER, DIMENSION(:), POINTER           :: lmax
    INTEGER, DIMENSION(:), POINTER           :: lmin
    INTEGER, DIMENSION(:), POINTER           :: npgf
    INTEGER                                  :: nset
    REAL(dp), DIMENSION(:,:), POINTER        :: zet
    INTEGER, DIMENSION(:), POINTER           :: nsgf
    INTEGER, DIMENSION(:,:), POINTER         :: first_sgf
    REAL(dp), DIMENSION(:,:), POINTER        :: sphi
    INTEGER                                  :: nsgf_total
  END TYPE

! *****************************************************************************
  TYPE hfx_basis_info_type
    INTEGER                                  :: max_set
    INTEGER                                  :: max_sgf
    INTEGER                                  :: max_am
  END TYPE

! *****************************************************************************
  TYPE hfx_screen_coeff_type
    REAL(dp)                                 :: x(2)
  END TYPE

! *****************************************************************************
!> \brief stores some data used in construction of Kohn-Sham matrix
!> \param potential_parameter stores information on the potential (1/r, erfc(wr)/r
!> \param screening_parameter stores screening infos such as epsilon
!> \param memory_parameter stores infos on memory used for in-core calculations
!> \param periodic_parameter stores information on how to apply pbc
!> \param load_balance_parameter contains infos for Monte Carlo simulated annealing
!> \param general_paramter at the moment stores the fraction of HF amount to be included
!> \param maxval_container stores the maxvals in compressed form
!> \param maxval_cache cache for maxvals in decompressed form
!> \param integral_containers 64 containers for compressed integrals
!> \param integral_caches 64 caches for decompressed integrals
!> \param neighbor_cells manages handling of periodic cells
!> \param distribution_energy stores information on parallelization of energy
!> \param distribution_forces stores information on parallelization of forces
!> \param initial_p stores the initial guess if requested
!> \param is_assoc_atomic_block reflects KS sparsity
!> \param number_of_p_entries Size of P matrix
!> \param n_rep_hf Number of HFX replicas
!> \param b_first_load_balance_x flag to indicate if it is enough just to update
!>        the distribution of the integrals
!> \param full_ks_x full ks matrices
!> \param lib libint type for eris
!> \param basis_info contains information for basis sets
!> \param screen_funct_coeffs_pgf pgf based near field screening coefficients
!> \param pair_dist_radii_pgf pgf based radii coefficients of pair distributions
!> \param screen_funct_coeffs_set set based near field screening coefficients
!> \param screen_funct_coeffs_kind kind based near field screening coefficients
!> \param screen_funct_is_initialized flag that indicates if the coefficients 
!>        have already been fitted
!> \par History
!>      11.2006 created [Manuel Guidon]
!>      02.2009 completely rewritten due to new screening
!> \author Manuel Guidon
! *****************************************************************************
  TYPE hfx_type
     TYPE(hfx_potential_type)                 :: potential_parameter
     TYPE(hfx_screening_type)                 :: screening_parameter
     TYPE(hfx_memory_type)                    :: memory_parameter
     TYPE(hfx_periodic_type)                  :: periodic_parameter
     TYPE(hfx_load_balance_type)              :: load_balance_parameter
     TYPE(hfx_general_type)                   :: general_parameter
     TYPE(hfx_container_type), POINTER        :: maxval_container, maxval_container_disk
     TYPE(hfx_cache_type)                     :: maxval_cache, maxval_cache_disk
     TYPE(hfx_container_type), POINTER, &
                                DIMENSION(:)  :: integral_containers, integral_containers_disk
     TYPE(hfx_cache_type)                     :: integral_caches(64), integral_caches_disk(64)

     TYPE(hfx_cell_type), DIMENSION(:), &
                POINTER                       :: neighbor_cells
     TYPE(hfx_distribution), DIMENSION(:),&
                              POINTER         :: distribution_energy
     TYPE(hfx_distribution), DIMENSION(:),&
                              POINTER         :: distribution_forces
     REAL(dp), DIMENSION(:,:,:,:), POINTER &
                                              :: initial_p
     INTEGER, DIMENSION(:,:), POINTER         :: is_assoc_atomic_block
     INTEGER                                  :: number_of_p_entries
     TYPE(hfx_basis_type), DIMENSION(:), &
                            POINTER           :: basis_parameter
     INTEGER                                  :: n_rep_hf
     LOGICAL                                  :: b_first_load_balance_energy, &
                                                 b_first_load_balance_forces
     REAL(dp), DIMENSION(:), POINTER          :: full_ks_alpha
     REAL(dp), DIMENSION(:), POINTER          :: full_ks_beta
     TYPE(lib_int)                            :: lib
     TYPE(hfx_basis_info_type)                :: basis_info
     TYPE(hfx_screen_coeff_type), &
          DIMENSION(:,:,:,:,:,:), POINTER     :: screen_funct_coeffs_pgf,&
                                                 pair_dist_radii_pgf
     TYPE(hfx_screen_coeff_type), &
          DIMENSION(:,:,:,:), POINTER         :: screen_funct_coeffs_set
     TYPE(hfx_screen_coeff_type), &
          DIMENSION(:,:), POINTER             :: screen_funct_coeffs_kind
     LOGICAL                                  :: screen_funct_is_initialized
  END TYPE hfx_type

! *****************************************************************************
!> \brief to build arrays of pointers
!> \param hfx_data the pointer to the hfx_type
!> \par History
!>      02.2008 created [mguidon]
!> \author Manuel Guidon
! *****************************************************************************
  TYPE hfx_p_type
     TYPE(hfx_type), POINTER :: hfx_data
  END TYPE hfx_p_type


  CONTAINS

! *****************************************************************************
!> \brief - This routine allocates and initializes all types in hfx_data
!> \param x_data contains all relevant data structures for hfx runs
!> \param hfx_section input section
!> \param natom , max_set: quantities that define some array sizes
!> \param error variable to control error logging, stopping,...
!>        see module cp_error_handling
!> \note
!>      - All POINTERS and ALLOCATABLES are allocated, even if their size is
!>        unknown at invocation time
!> \par History
!>      09.2007 created [Manuel Guidon]
!> \author Manuel Guidon
! *****************************************************************************
  SUBROUTINE hfx_create(x_data,para_env, hfx_section,natom,atomic_kind_set, &
                        dft_control, cell, error)
    TYPE(hfx_type), DIMENSION(:, :), POINTER :: x_data
    TYPE(cp_para_env_type)                   :: para_env
    TYPE(section_vals_type), POINTER         :: hfx_section
    INTEGER, INTENT(IN)                      :: natom
    TYPE(atomic_kind_type), DIMENSION(:), &
      POINTER                                :: atomic_kind_set
    TYPE(dft_control_type), POINTER          :: dft_control
    TYPE(cell_type), POINTER                 :: cell
    TYPE(cp_error_type), INTENT(inout)       :: error

    CHARACTER(LEN=*), PARAMETER :: routineN = 'hfx_create', &
      routineP = moduleN//':'//routineN

    CHARACTER(LEN=512)                       :: error_msg
    CHARACTER(LEN=default_path_length)       :: char_val, filename
    INTEGER :: handle, i, i_thread, ikind, int_val, irep, max_set, n_rep_hf, &
      n_threads, nkind, pbc_shells, stat, storage_id
    LOGICAL                                  :: do_periodic, exist, failure, &
                                                logic_val
    REAL(dp)                                 :: omega, real_val
    TYPE(atomic_kind_type), POINTER          :: atom_kind
    TYPE(gto_basis_set_type), POINTER        :: orb_basis_a
    TYPE(hfx_type), POINTER                  :: actual_x_data
    TYPE(section_vals_type), POINTER         :: hf_pbc_section, hf_sub_section

!$  INTEGER                                  :: omp_get_max_threads

    CALL timeset(routineN,handle)

    failure = .FALSE.
    CALL cite_reference(Guidon2008)
    !! There might be 2 hf sections
    CALL section_vals_get(hfx_section,n_repetition=n_rep_hf,error=error)
    n_threads = 1
!$  n_threads = omp_get_max_threads() 
    ALLOCATE(x_data(n_rep_hf,n_threads),STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    DO i_thread=1,n_threads
      DO irep=1,n_rep_hf
        actual_x_data => x_data(irep,i_thread)
        !! Get data from input file
        !!
        !! GENERAL params
        CALL section_vals_val_get(hfx_section, "FRACTION", r_val=real_val, i_rep_section=irep,error=error)
        actual_x_data%general_parameter%fraction = real_val
        actual_x_data%n_rep_hf = n_rep_hf  
       
        CALL section_vals_val_get(hfx_section, "TREAT_LSD_IN_CORE", l_val=logic_val, i_rep_section=irep,error=error)
        actual_x_data%general_parameter%treat_lsd_in_core = logic_val

  
        !! MEMORY section
        hf_sub_section => section_vals_get_subs_vals(hfx_section,"MEMORY",i_rep_section=irep, error=error)
        CALL parse_memory_section(actual_x_data%memory_parameter, hf_sub_section, storage_id, i_thread,&
             n_threads, para_env, irep, skip_disk=.FALSE., error=error)

        !! PERIODIC section
        hf_sub_section => section_vals_get_subs_vals(hfx_section,"PERIODIC", i_rep_section=irep, error=error)
        CALL section_vals_val_get(hf_sub_section, "NUMBER_OF_SHELLS", i_val=int_val, error=error)
        actual_x_data%periodic_parameter%number_of_shells = int_val
        actual_x_data%periodic_parameter%mode = int_val
        CALL get_cell(cell=cell,periodic=actual_x_data%periodic_parameter%perd)
        IF( SUM(actual_x_data%periodic_parameter%perd) == 0 ) THEN
          actual_x_data%periodic_parameter%do_periodic = .FALSE.
        ELSE
          actual_x_data%periodic_parameter%do_periodic = .TRUE.
        END IF
 
        !! SCREENING section
        hf_sub_section => section_vals_get_subs_vals(hfx_section,"SCREENING",i_rep_section=irep, error=error)
        CALL section_vals_val_get(hf_sub_section, "EPS_SCHWARZ", r_val=real_val, error=error)
        actual_x_data%screening_parameter%eps_schwarz = real_val
        CALL section_vals_val_get(hf_sub_section, "EPS_SCHWARZ_FORCES", r_val=real_val, error=error)
        actual_x_data%screening_parameter%eps_schwarz_forces = real_val 
        CALL section_vals_val_get(hf_sub_section, "SCREEN_P_FORCES", l_val=logic_val, error=error)
        actual_x_data%screening_parameter%do_p_screening_forces = logic_val 
        CALL section_vals_val_get(hf_sub_section, "SCREEN_ON_INITIAL_P", l_val=logic_val, error=error)
        actual_x_data%screening_parameter%do_initial_p_screening = logic_val 
        actual_x_data%screen_funct_is_initialized = .FALSE.
    
        !! INTERACTION_POTENTIAL section
        hf_sub_section => section_vals_get_subs_vals(hfx_section,"INTERACTION_POTENTIAL",i_rep_section=irep,error=error)
        CALL section_vals_val_get(hf_sub_section, "POTENTIAL_TYPE", i_val=int_val, error=error)
        actual_x_data%potential_parameter%potential_type = int_val
        CALL section_vals_val_get(hf_sub_section, "OMEGA", r_val=real_val, error=error)
        actual_x_data%potential_parameter%omega = real_val 
        CALL section_vals_val_get(hf_sub_section, "SCALE_COULOMB", r_val=real_val, error=error)
        actual_x_data%potential_parameter%scale_coulomb = real_val
        CALL section_vals_val_get(hf_sub_section, "SCALE_LONGRANGE", r_val=real_val, error=error)
        actual_x_data%potential_parameter%scale_longrange = real_val
        CALL section_vals_val_get(hf_sub_section, "SCALE_GAUSSIAN", r_val=real_val, error=error)
        actual_x_data%potential_parameter%scale_gaussian = real_val
        IF (actual_x_data%potential_parameter%potential_type == do_hfx_potential_truncated) THEN
          CALL section_vals_val_get(hf_sub_section, "CUTOFF_RADIUS", r_val=real_val, error=error)
          actual_x_data%potential_parameter%cutoff_radius = real_val
          CALL section_vals_val_get(hf_sub_section, "T_C_G_DATA", c_val=char_val, error=error)
          CALL compress(char_val,.TRUE.)
          ! ** Check if file is there
          INQUIRE(FILE=char_val,exist=exist) 
          IF ( .NOT. exist ) THEN
            WRITE(error_msg,'(A,A,A)') "Truncated hfx calculation requested. The file containing "//&
                 "the data could not be found at ", TRIM(char_val), " Please check T_C_G_DATA "//&
                 "in the INTERCATION_POTENTIAL section"
            CALL cp_assert( .FALSE. , cp_failure_level,cp_assertion_failed,routineP,&
                 error_msg , &
                 error,&
                 only_ionode=.TRUE.)
          ELSE
            actual_x_data%potential_parameter%filename = char_val
          END IF
        END IF
    
        !! LOAD_BALANCE section
        hf_sub_section => section_vals_get_subs_vals(hfx_section,"LOAD_BALANCE",i_rep_section=irep, error=error)
        CALL section_vals_val_get(hf_sub_section, "NBINS", i_val=int_val, error=error)
        actual_x_data%load_balance_parameter%nbins = MAX(1,int_val)

        CALL section_vals_val_get(hf_sub_section, "BLOCK_SIZE", i_val=int_val, error=error)
        ! negative values ask for a computed default
        IF (int_val<=0) THEN
           ! this gives a reasonable number of blocks for binning, yet typically results in blocking.
           int_val = CEILING( 0.1_dp * natom / &
                 REAL(actual_x_data%load_balance_parameter%nbins * n_threads * para_env%num_pe ,KIND=dp)**(0.25_dp))
        ENDIF
        ! at least 1 atom per block, and avoid overly large blocks
        actual_x_data%load_balance_parameter%block_size = MIN(max_atom_block, MAX(1, int_val))
   
        !! BASIS parameter
        nkind = SIZE(atomic_kind_set,1)
        ALLOCATE(actual_x_data%basis_parameter(nkind), STAT=stat)
        CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
        max_set = 0
        DO ikind = 1,nkind
            NULLIFY(atom_kind)
            atom_kind => atomic_kind_set(ikind)
            IF( .NOT.dft_control%do_wfn_fitting ) THEN
              CALL get_atomic_kind(atomic_kind=atom_kind,&
                                   orb_basis_set=orb_basis_a)
              CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set,&
                                       maxsgf=actual_x_data%basis_info%max_sgf,&
                                       maxnset=actual_x_data%basis_info%max_set,&
                                       maxlgto=actual_x_data%basis_info%max_am)

            ELSE
              CALL get_atomic_kind(atomic_kind=atom_kind,&
                                   aux_fit_basis_set=orb_basis_a)
              CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set,&
                                       maxsgf=actual_x_data%basis_info%max_sgf,&
                                       maxnset=actual_x_data%basis_info%max_set,&
                                       maxlgto=actual_x_data%basis_info%max_am,&
                                       basis_set_id=use_aux_fit_basis_set)
            END IF
            max_set = MAX(max_set,actual_x_data%basis_info%max_set)
            CALL get_gto_basis_set(gto_basis_set=orb_basis_a, &
                                   lmax=actual_x_data%basis_parameter(ikind)%lmax, &
                                   lmin=actual_x_data%basis_parameter(ikind)%lmin, &
                                   npgf=actual_x_data%basis_parameter(ikind)%npgf, &
                                   nset=actual_x_data%basis_parameter(ikind)%nset, &
                                   zet=actual_x_data%basis_parameter(ikind)%zet, &
                                   nsgf_set=actual_x_data%basis_parameter(ikind)%nsgf, &
                                   first_sgf=actual_x_data%basis_parameter(ikind)%first_sgf, &
                                   sphi=actual_x_data%basis_parameter(ikind)%sphi, &
                                   nsgf=actual_x_data%basis_parameter(ikind)%nsgf_total) 
        END DO
    
        IF(actual_x_data%periodic_parameter%do_periodic) THEN
          hf_pbc_section => section_vals_get_subs_vals(hfx_section,"PERIODIC",i_rep_section=irep, error=error)
          CALL section_vals_val_get(hf_pbc_section,"NUMBER_OF_SHELLS",i_val=pbc_shells,error=error)
          ALLOCATE(actual_x_data%neighbor_cells(1),STAT=stat)
          CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
          CALL hfx_create_neighbor_cells(actual_x_data,pbc_shells, cell, hfx_section, i_thread, error=error)
        END IF
     
        !! ** This guy is allocated on the master thread only 
        IF (i_thread == 1 ) THEN 
          ALLOCATE(actual_x_data%is_assoc_atomic_block(natom,natom),STAT=stat)
          CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
        END IF
    
        ALLOCATE(actual_x_data%distribution_forces(1),STAT=stat)
        CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
        ALLOCATE(actual_x_data%distribution_energy(1),STAT=stat)
        CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    
        IF(actual_x_data%screening_parameter%do_initial_p_screening .OR. &
           actual_x_data%screening_parameter%do_p_screening_forces) THEN
          !! ** This guy is allocated on the master thread only 
          IF (i_thread == 1 ) THEN
            ALLOCATE(actual_x_data%initial_p(max_set,max_set,natom,natom),STAT=stat)
            CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
          END IF
        END IF
    
        actual_x_data%maxval_cache%element_counter = 1
        ALLOCATE(actual_x_data%maxval_container,STAT=stat)
        CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
        ALLOCATE(actual_x_data%maxval_container%first,STAT=stat)
        CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
        actual_x_data%maxval_container%first%prev => NULL()
        actual_x_data%maxval_container%first%next => NULL()
        actual_x_data%maxval_container%current => actual_x_data%maxval_container%first
        actual_x_data%maxval_container%current%data = 0
        actual_x_data%maxval_container%element_counter = 1
    
        ALLOCATE(actual_x_data%integral_containers(64), STAT=stat)
        CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
        DO i=1,64
          actual_x_data%integral_caches(i)%element_counter = 1
          actual_x_data%integral_caches(i)%data = 0
          CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
          ALLOCATE(actual_x_data%integral_containers(i)%first,STAT=stat)
          CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
          actual_x_data%integral_containers(i)%first%prev => NULL()
          actual_x_data%integral_containers(i)%first%next => NULL()
          actual_x_data%integral_containers(i)%current => actual_x_data%integral_containers(i)%first
          actual_x_data%integral_containers(i)%current%data = 0
          actual_x_data%integral_containers(i)%element_counter = 1
        END DO

        actual_x_data%maxval_cache_disk%element_counter = 1
        ALLOCATE(actual_x_data%maxval_container_disk,STAT=stat)
        CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
        ALLOCATE(actual_x_data%maxval_container_disk%first,STAT=stat)
        CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
        actual_x_data%maxval_container_disk%first%prev => NULL()
        actual_x_data%maxval_container_disk%first%next => NULL()
        actual_x_data%maxval_container_disk%current => actual_x_data%maxval_container%first
        actual_x_data%maxval_container_disk%current%data = 0
        actual_x_data%maxval_container_disk%element_counter = 1
        actual_x_data%maxval_container_disk%file_counter = 1
        actual_x_data%maxval_container_disk%desc = 'Max_'
        actual_x_data%maxval_container_disk%unit = -1
        WRITE(actual_x_data%maxval_container_disk%filename,'(A,I0,A,A,A)') TRIM(actual_x_data%memory_parameter%storage_location), &
                                                           storage_id,"_",actual_x_data%maxval_container_disk%desc, "6"
        CALL compress(actual_x_data%maxval_container_disk%filename, .TRUE.)
        ALLOCATE(actual_x_data%integral_containers_disk(64), STAT=stat)
        CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
        DO i=1,64
          actual_x_data%integral_caches_disk(i)%element_counter = 1
          actual_x_data%integral_caches_disk(i)%data = 0
          CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
          ALLOCATE(actual_x_data%integral_containers_disk(i)%first,STAT=stat)
          CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
          actual_x_data%integral_containers_disk(i)%first%prev => NULL()
          actual_x_data%integral_containers_disk(i)%first%next => NULL()
          actual_x_data%integral_containers_disk(i)%current => actual_x_data%integral_containers(i)%first
          actual_x_data%integral_containers_disk(i)%current%data = 0
          actual_x_data%integral_containers_disk(i)%element_counter = 1
          actual_x_data%integral_containers_disk(i)%file_counter = 1
          actual_x_data%integral_containers_disk(i)%desc = 'Int_'
          actual_x_data%integral_containers_disk(i)%unit = -1
          WRITE(actual_x_data%integral_containers_disk(i)%filename,'(A,I0,A,A,I0)')&
                                             TRIM(actual_x_data%memory_parameter%storage_location), &
                                             storage_id,"_",actual_x_data%integral_containers_disk(i)%desc, i
          CALL compress(actual_x_data%integral_containers_disk(i)%filename, .TRUE.)
        END DO

        actual_x_data%b_first_load_balance_energy = .TRUE.
        actual_x_data%b_first_load_balance_forces = .TRUE.
      END DO
    END DO

    DO irep=1,n_rep_hf
      actual_x_data => x_data(irep,1)
      CALL hfx_print_info(actual_x_data, hfx_section, irep, error=error)
    END DO
    CALL timestop(handle)
  END SUBROUTINE hfx_create 

! *****************************************************************************
!> \brief - Parses the memory section
! *****************************************************************************
  SUBROUTINE parse_memory_section(memory_parameter, hf_sub_section, storage_id,&
       i_thread, n_threads, para_env, irep, skip_disk, error)
    TYPE(hfx_memory_type)                    :: memory_parameter
    TYPE(section_vals_type), POINTER         :: hf_sub_section
    INTEGER, INTENT(OUT), OPTIONAL           :: storage_id
    INTEGER, INTENT(IN), OPTIONAL            :: i_thread, n_threads
    TYPE(cp_para_env_type), OPTIONAL         :: para_env
    INTEGER, INTENT(IN), OPTIONAL            :: irep
    LOGICAL, INTENT(IN)                      :: skip_disk
    TYPE(cp_error_type), INTENT(inout)       :: error

    CHARACTER(LEN=*), PARAMETER :: routineN = 'parse_memory_section', &
      routineP = moduleN//':'//routineN

    CHARACTER(LEN=512)                       :: error_msg
    CHARACTER(LEN=default_path_length)       :: char_val, filename, orig_wd
    INTEGER                                  :: int_val, stat
    LOGICAL                                  :: check, failure
    REAL(dp)                                 :: real_val

    failure = .FALSE.
    check = (PRESENT(storage_id).EQV.PRESENT(i_thread)) .AND.&
            (PRESENT(storage_id).EQV.PRESENT(n_threads)).AND.&
            (PRESENT(storage_id).EQV.PRESENT(para_env)) .AND.&
            (PRESENT(storage_id).EQV.PRESENT(irep))
    CPPostcondition(check,cp_failure_level,routineP,error,failure)
    
    ! Memory Storage
    CALL section_vals_val_get(hf_sub_section, "MAX_MEMORY", i_val=int_val, error=error)
    memory_parameter%max_memory = int_val
    memory_parameter%max_compression_counter = int_val*1024_int_8*128_int_8
    CALL section_vals_val_get(hf_sub_section, "EPS_STORAGE", r_val=real_val, error=error)
    memory_parameter%eps_storage_scaling = real_val
    IF(int_val == 0) THEN
       memory_parameter%do_all_on_the_fly = .TRUE.
    ELSE
       memory_parameter%do_all_on_the_fly = .FALSE.
    END IF
    memory_parameter%cache_size = CACHE_SIZE
    memory_parameter%actual_memory_usage = 1

    ! Disk Storage
    IF (.NOT.skip_disk) THEN
       memory_parameter%actual_memory_usage_disk = 1
       CALL section_vals_val_get(hf_sub_section, "MAX_DISK_SPACE", i_val=int_val, error=error)
       memory_parameter%max_compression_counter_disk = int_val*1024_int_8*128_int_8
       IF( int_val == 0 ) THEN
          memory_parameter%do_disk_storage = .FALSE.
       ELSE
          memory_parameter%do_disk_storage = .TRUE.
       END IF
       CALL section_vals_val_get(hf_sub_section, "STORAGE_LOCATION", c_val=char_val, error=error)
       CALL compress(char_val,.TRUE.)
       !! Add ending / if necessary
       
       IF( SCAN(char_val,"/",.TRUE.) /= LEN_TRIM(char_val)) THEN
          WRITE(filename,'(A,A)') TRIM(char_val),"/"
          CALL compress(filename)
       ELSE
          filename = TRIM(char_val)
       END IF
       CALL compress(filename,.TRUE.)
       
       !! quickly check if we can write on storage_location
       CALL m_getcwd(orig_wd)
       CALL m_chdir(TRIM(filename),stat)
       IF(stat/=0) THEN
          WRITE(error_msg,'(A,A,A)') "Request for disk storage failed due to unknown error while writing to ",&
               TRIM(filename), ". Please check STORAGE_LOCATION"
          CALL cp_assert( .FALSE. , cp_failure_level,cp_assertion_failed,routineP,&
               error_msg , &
               error,failure)
       END IF
       CALL m_chdir(orig_wd,stat)

       memory_parameter%storage_location = filename
       CALL compress(memory_parameter%storage_location, .TRUE.)
    ELSE
       memory_parameter%do_disk_storage = .FALSE.
    END IF
    IF (PRESENT(storage_id)) THEN
       storage_id = (irep-1)*para_env%num_pe*n_threads + para_env%mepos * n_threads + i_thread-1
    END IF
  END SUBROUTINE parse_memory_section

! *****************************************************************************
!> \brief - This routine deallocates all data structures
!> \param x_data contains all relevant data structures for hfx runs
!> \param error variable to control error logging, stopping,...
!>        see module cp_error_handling
!> \par History
!>      09.2007 created [Manuel Guidon]
!> \author Manuel Guidon
! *****************************************************************************
  SUBROUTINE hfx_release(x_data, error)
    TYPE(hfx_type), DIMENSION(:, :), POINTER :: x_data
    TYPE(cp_error_type), INTENT(inout)       :: error

    CHARACTER(LEN=*), PARAMETER :: routineN = 'hfx_release', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: i, i_thread, irep, n_rep_hf, &
                                                n_threads, stat
    LOGICAL                                  :: failure
    TYPE(hfx_type), POINTER                  :: actual_x_data

!$  INTEGER                                  :: omp_get_max_threads


    failure = .FALSE.
    !! There might be 2 hf sections
    n_rep_hf = x_data(1,1)%n_rep_hf
    n_threads = 1
!$  n_threads = omp_get_max_threads()

    IF(x_data(1,1)%potential_parameter%potential_type == do_hfx_potential_truncated) THEN
      CALL free()
    END IF
    DO i_thread= 1,n_threads
      DO irep=1,n_rep_hf
        actual_x_data => x_data(irep,i_thread)
        IF( actual_x_data%periodic_parameter%do_periodic ) THEN
          DEALLOCATE(actual_x_data%neighbor_cells,STAT=stat)
          CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
        END IF
        DEALLOCATE(actual_x_data%distribution_energy,STAT=stat)
        CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
        DEALLOCATE(actual_x_data%distribution_forces,STAT=stat)
        CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    
        IF(actual_x_data%screening_parameter%do_initial_p_screening .OR. &
           actual_x_data%screening_parameter%do_p_screening_forces) THEN
          IF( i_thread == 1 ) THEN
            DEALLOCATE(actual_x_data%initial_p,STAT=stat)
            CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
          END IF
        END IF
        IF( i_thread == 1 ) THEN 
          DEALLOCATE(actual_x_data%is_assoc_atomic_block,STAT=stat)
          CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
        END IF
    
        !! BASIS parameter
        DEALLOCATE(actual_x_data%basis_parameter,STAT=stat)
        CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    
        !! Deallocate containers
        CALL hfx_init_container(6, actual_x_data%maxval_container, actual_x_data%memory_parameter%actual_memory_usage, &
                                 .FALSE., error)
        DEALLOCATE(actual_x_data%maxval_container%first,STAT=stat)
        CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
        DEALLOCATE(actual_x_data%maxval_container,STAT=stat)
        CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    
        DO i=1,64   
          CALL hfx_init_container(i, actual_x_data%integral_containers(i), actual_x_data%memory_parameter%actual_memory_usage, &
                                   .FALSE., error)
          DEALLOCATE(actual_x_data%integral_containers(i)%first,STAT=stat)
          CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
        END DO
        DEALLOCATE(actual_x_data%integral_containers,STAT=stat)
        CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)

        !! Deallocate containers
        CALL hfx_init_container(6, actual_x_data%maxval_container_disk, &
                                 actual_x_data%memory_parameter%actual_memory_usage_disk, &
                                 .FALSE., error)
        IF( actual_x_data%memory_parameter%do_disk_storage) THEN
          CALL close_file(unit_number=actual_x_data%maxval_container_disk%unit,file_status="DELETE")
        END IF
        DEALLOCATE(actual_x_data%maxval_container_disk%first,STAT=stat)
        CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
        DEALLOCATE(actual_x_data%maxval_container_disk,STAT=stat)
        CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    
        DO i=1,64   
          CALL hfx_init_container(i, actual_x_data%integral_containers_disk(i), &
                                   actual_x_data%memory_parameter%actual_memory_usage_disk, &
                                   .FALSE., error)
          IF( actual_x_data%memory_parameter%do_disk_storage) THEN
            CALL close_file(unit_number=actual_x_data%integral_containers_disk(i)%unit,file_status="DELETE")
          END IF
          DEALLOCATE(actual_x_data%integral_containers_disk(i)%first,STAT=stat)
          CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
        END DO
        DEALLOCATE(actual_x_data%integral_containers_disk,STAT=stat)
        CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)


        ! ** screening functions
        IF(actual_x_data%screen_funct_is_initialized) THEN
          DEALLOCATE(actual_x_data%screen_funct_coeffs_set,STAT=stat)
          CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
          DEALLOCATE(actual_x_data%screen_funct_coeffs_kind,STAT=stat)
          CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
          DEALLOCATE(actual_x_data%pair_dist_radii_pgf,STAT=stat)
          CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
          DEALLOCATE(actual_x_data%screen_funct_coeffs_pgf,STAT=stat)
          CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
          actual_x_data%screen_funct_is_initialized = .FALSE. 
        END IF

      END DO

    END DO
    DEALLOCATE(x_data,STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
  END SUBROUTINE hfx_release

! *****************************************************************************
!> \brief - This routine computes the neighbor cells that are taken into account
!>        in periodic runs
!> \param x_data contains all relevant data structures for hfx runs
!> \param pbc_shells number of shells taken into account
!> \param cell cell
!> \param hfx_section HFX input section
!> \param i_thread current thread ID
!> \param error variable to control error logging, stopping,...
!>        see module cp_error_handling
!> \par History
!>      09.2007 created [Manuel Guidon]
!> \author Manuel Guidon
! *****************************************************************************
  SUBROUTINE hfx_create_neighbor_cells(x_data, pbc_shells, cell, hfx_section, i_thread, &
                                       error)
    TYPE(hfx_type), POINTER                  :: x_data
    INTEGER, INTENT(INOUT)                   :: pbc_shells
    TYPE(cell_type), POINTER                 :: cell
    TYPE(section_vals_type), POINTER         :: hfx_section
    INTEGER, INTENT(IN)                      :: i_thread
    TYPE(cp_error_type), INTENT(inout)       :: error

    CHARACTER(LEN=*), PARAMETER :: routineN = 'hfx_create_neighbor_cells', &
      routineP = moduleN//':'//routineN

    CHARACTER(LEN=512)                       :: error_msg
    CHARACTER(LEN=64)                        :: char_nshells
    INTEGER :: act_idx, i, ikind, ipgf, iset, j, jkind, jpgf, jset, k, l, &
      m(3), max_cells_old, ncells_old, nseta, nsetb, nshells, nshells_old, &
      stat, total_number_of_cells
    INTEGER, DIMENSION(:), POINTER           :: la_max, lb_max, npgfa, npgfb
    LOGICAL                                  :: failure, max_shell_found
    REAL(dp)                                 :: dist, dist_min, l_min, r(3), &
                                                R1, R_max, s(3), x, y, z, &
                                                Zeta1
    REAL(dp), DIMENSION(:, :), POINTER       :: zeta, zetb
    TYPE(cp_logger_type), POINTER            :: logger
    TYPE(hfx_cell_type), ALLOCATABLE, &
      DIMENSION(:)                           :: tmp_neighbor_cells

    NULLIFY(logger)
    logger => cp_error_get_logger(error)

    failure = .FALSE.
    total_number_of_cells = 0

    ! ** Check some settings
    IF( i_thread == 1 ) THEN
      IF( x_data%potential_parameter%potential_type /= do_hfx_potential_truncated) THEN
        CALL cp_assert(.FALSE.,cp_warning_level,cp_assertion_failed,routineP,&
                   "Periodic Hartree Fock calculation requested without use "//&
                   "of a truncated potential. This may lead to unphysical total energies. "//&
                   "Use a truncated  potential to avoid possible problems. "//&
                   CPSourceFileRef,&
                   only_ionode=.TRUE.)
      ELSE
        l_min = MIN( plane_distance(1,0,0,cell), &
                     plane_distance(0,1,0,cell), &
                     plane_distance(0,0,1,cell) )
        l_min = 0.5_dp * l_min
        IF( x_data%potential_parameter%cutoff_radius >= l_min ) THEN
          CALL cp_assert(.FALSE.,cp_warning_level,cp_assertion_failed,routineP,&
                   "Periodic Hartree Fock calculation requested with use "//&
                   "of a truncated potential. The cutoff radius is larger than half "//&
                   "the minimal cell dimension. This may lead to unphysical "//&
                   "total energies. Reduce the cutoff radius in order to avoid "//&
                   "possible problems. "//&
                   CPSourceFileRef,&
                   only_ionode=.TRUE.)
        END IF
      END IF
    END IF

    SELECT CASE(x_data%potential_parameter%potential_type)
    CASE( do_hfx_potential_truncated )
      R_max = 0.0_dp
      DO ikind=1,SIZE(x_data%basis_parameter)
        la_max => x_data%basis_parameter(ikind)%lmax
        zeta => x_data%basis_parameter(ikind)%zet
        nseta = x_data%basis_parameter(ikind)%nset
        npgfa => x_data%basis_parameter(ikind)%npgf
        DO jkind=1,SIZE(x_data%basis_parameter)
          lb_max => x_data%basis_parameter(jkind)%lmax
          zetb => x_data%basis_parameter(jkind)%zet
          nsetb = x_data%basis_parameter(jkind)%nset
          npgfb => x_data%basis_parameter(jkind)%npgf
          DO iset=1,nseta
            DO jset=1,nsetb
              DO ipgf=1,npgfa(iset)
                DO jpgf=1,npgfb(jset)
                  Zeta1 = zeta(ipgf,iset) + zetb(jpgf,jset)
                  R1 = 1.0_dp/SQRT(Zeta1) * mul_fact(la_max(iset)+lb_max(jset)) * &
                       SQRT(-LOG(x_data%screening_parameter%eps_schwarz))
                  R_max = MAX(R1,R_max)
                END DO
              END DO
            END DO
          END DO
        END DO
      END DO
  
      R_max = 2.0_dp*R_max + x_data%potential_parameter%cutoff_radius
 
      max_shell_found = .FALSE.
      nshells = -1
      nshells_old = -1
      ncells_old = -1
      max_cells_old = -1
      DO WHILE( .NOT. max_shell_found ) 
        total_number_of_cells = 0
        nshells = nshells + 1
        DO i = 0,nshells
          total_number_of_cells = total_number_of_cells + count_cells_perd(i,x_data%periodic_parameter%perd)
        END DO
        IF( total_number_of_cells == max_cells_old ) THEN
          CYCLE
        END IF
        max_cells_old = total_number_of_cells
        DEALLOCATE(x_data%neighbor_cells,STAT=stat)
        CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    
        ALLOCATE(x_data%neighbor_cells(total_number_of_cells), STAT=stat)
        CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
        ALLOCATE(tmp_neighbor_cells(total_number_of_cells), STAT=stat)
        CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
   
        m = 0
        i = 1
        DO WHILE(SUM(m**2)<=nshells)
          x_data%neighbor_cells(i)%cell = REAL(m,dp)
          CALL next_image_cell_perd(m,x_data%periodic_parameter%perd)
          i=i+1
        END DO
        act_idx = 1
        DO i = 1,SIZE(x_data%neighbor_cells)
          dist_min = HUGE(dist_min)
          DO j=0,1
            x = -1.0_dp/2.0_dp + j*1.0_dp
            DO k=0,1
              y = -1.0_dp/2.0_dp + k*1.0_dp 
              DO l=0,1 
                z = -1.0_dp/2.0_dp + l*1.0_dp
                s(1) = x+x_data%neighbor_cells(i)%cell(1)
                s(2) = y+x_data%neighbor_cells(i)%cell(2)
                s(3) = z+x_data%neighbor_cells(i)%cell(3)
                CALL scaled_to_real(r,s,cell)
                dist = SQRT(SUM(r**2))
                dist_min = MIN(dist,dist_min)
              END DO
            END DO
          END DO
          IF( dist_min <= R_max) THEN
            tmp_neighbor_cells(act_idx) = x_data%neighbor_cells(i)
            act_idx = act_idx + 1
          END IF
        END DO
     
        total_number_of_cells = act_idx-1
        IF( ncells_old == total_number_of_cells) max_shell_found = .TRUE.
        nshells_old = nshells
        ncells_old = total_number_of_cells
        IF( .NOT. max_shell_found ) THEN
          DEALLOCATE(tmp_neighbor_cells,STAT=stat)
          CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
        END IF
      END DO
      DEALLOCATE(x_data%neighbor_cells,STAT=stat)
      CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
      IF(x_data%periodic_parameter%number_of_shells == do_hfx_auto_shells ) THEN 
        ! if only the central unit cell is needed we need to take care manually
        ! to avoid undefined cell info
        IF(total_number_of_cells == 0) THEN
            total_number_of_cells = 1
            tmp_neighbor_cells(1)%cell=0
            tmp_neighbor_cells(1)%cell_r=0
        ENDIF
        ALLOCATE(x_data%neighbor_cells(total_number_of_cells), STAT=stat)
        CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
        x_data%neighbor_cells(1:total_number_of_cells) = tmp_neighbor_cells(1:total_number_of_cells)
        pbc_shells = nshells - 1
      ELSE
        IF( x_data%periodic_parameter%number_of_shells < nshells - 1) THEN
          IF( i_thread == 1) THEN
            WRITE(char_nshells,'(I3)') nshells-1
            WRITE(error_msg,'(A,A,A)') "Periodic Hartree Fock calculation requested with use "//&
                     "of a truncated potential. The number of shells to be considered "//&
                     "might be too small. CP2K conservatively estimates to need "//TRIM(char_nshells)//" shells. "//&
                     "Please carefully check if you get converged results."

            CALL cp_assert(.FALSE.,cp_warning_level,cp_assertion_failed,routineP,&
                           error_msg,&
                           error,&
                           only_ionode=.TRUE.)
          END IF
        END IF
        total_number_of_cells = 0
        DO i = 0,x_data%periodic_parameter%number_of_shells
          total_number_of_cells = total_number_of_cells + count_cells_perd(i,x_data%periodic_parameter%perd)
        END DO
        ALLOCATE(x_data%neighbor_cells(total_number_of_cells), STAT=stat)
        CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
        m = 0
        i = 1
        DO WHILE(SUM(m**2)<=x_data%periodic_parameter%number_of_shells)
           x_data%neighbor_cells(i)%cell = REAL(m,dp)
           CALL next_image_cell_perd(m,x_data%periodic_parameter%perd)
           i=i+1
        ENDDO
      END IF
      DEALLOCATE(tmp_neighbor_cells,STAT=stat)
      CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    CASE DEFAULT
      total_number_of_cells = 0
      IF( pbc_shells == -1 ) pbc_shells = 0
      DO i = 0,pbc_shells
        total_number_of_cells = total_number_of_cells + count_cells_perd(i,x_data%periodic_parameter%perd)
      END DO
      DEALLOCATE(x_data%neighbor_cells,STAT=stat)
      CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)

      ALLOCATE(x_data%neighbor_cells(total_number_of_cells), STAT=stat)
      CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)

      m = 0
      i = 1
      DO WHILE(SUM(m**2)<=pbc_shells)
         x_data%neighbor_cells(i)%cell = REAL(m,dp)
         CALL next_image_cell_perd(m,x_data%periodic_parameter%perd)
         i=i+1
      ENDDO
    END SELECT

    ! ** Transform into real coord
    DO i=1,SIZE(x_data%neighbor_cells)
      r = 0.0_dp
      x_data%neighbor_cells(i)%cell_r(:) = 0.0_dp
      s = x_data%neighbor_cells(i)%cell(:)
      CALL scaled_to_real(x_data%neighbor_cells(i)%cell_r,s,cell)
    END DO

    x_data%periodic_parameter%number_of_shells = pbc_shells
  END SUBROUTINE hfx_create_neighbor_cells

! *****************************************************************************
!> \brief - This routine deletes all list entries in a container in order to
!>        deallocate the memory.
!> \param container container that contains the compressed elements
!> \param error variable to control error logging, stopping,...
!>        see module cp_error_handling
!> \par History
!>      10.2007 created [Manuel Guidon]
!> \author Manuel Guidon
! *****************************************************************************
  SUBROUTINE hfx_init_container(nbits, container, memory_usage, do_disk_storage, error)
    INTEGER                                  :: nbits
    TYPE(hfx_container_type)                 :: container
    INTEGER                                  :: memory_usage
    LOGICAL                                  :: do_disk_storage
    TYPE(cp_error_type), INTENT(inout)       :: error

    CHARACTER(LEN=*), PARAMETER :: routineN = 'hfx_init_container', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: stat
    LOGICAL                                  :: failure
    TYPE(hfx_container_node), POINTER        :: current, next

    failure =.FALSE.

    !! DEALLOCATE memory

    current => container%first
    DO WHILE(ASSOCIATED(current))
      next => current%next
      DEALLOCATE(current)
      current=>next
    END DO

    !! Allocate first list entry, init members
    ALLOCATE(container%first,STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    container%first%prev => NULL()
    container%first%next => NULL()
    container%current => container%first
    container%current%data = 0
    container%element_counter = 1
    memory_usage = 1

    IF( do_disk_storage) THEN
      !! close the file, if this is no the first time
      IF(container%unit /= -1 ) THEN
        CALL close_file(unit_number=container%unit)
      END IF
      CALL open_file(file_name=TRIM(container%filename),file_status="UNKNOWN",file_form="UNFORMATTED",file_action="WRITE",&
                     unit_number=container%unit)
    END IF

  END SUBROUTINE hfx_init_container


! *****************************************************************************
!> \brief - This routine stores the data obtained from the load balance routine 
!>        for the energy
!> \param ptr_to_distr_energy contains data to store
!> \param x_data contains all relevant data structures for hfx runs
!> \param error variable to control error logging, stopping,...
!>        see module cp_error_handling
!> \par History
!>      09.2007 created [Manuel Guidon]
!> \author Manuel Guidon
! *****************************************************************************
  SUBROUTINE hfx_set_distr_energy(ptr_to_distr,x_data, error)
    TYPE(hfx_distribution), DIMENSION(:), &
      POINTER                                :: ptr_to_distr
    TYPE(hfx_type), POINTER                  :: x_data
    TYPE(cp_error_type), INTENT(inout)       :: error

    CHARACTER(LEN=*), PARAMETER :: routineN = 'hfx_set_distr_energy', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: stat
    LOGICAL                                  :: failure

    DEALLOCATE(x_data%distribution_energy,STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)

    ALLOCATE(x_data%distribution_energy(SIZE(ptr_to_distr)),STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    x_data%distribution_energy = ptr_to_distr

  END SUBROUTINE hfx_set_distr_energy

! *****************************************************************************
!> \brief - This routine stores the data obtained from the load balance routine 
!>        for the forces
!> \param ptr_to_distr_energy contains data to store
!> \param x_data contains all relevant data structures for hfx runs
!> \param error variable to control error logging, stopping,...
!>        see module cp_error_handling
!> \par History
!>      09.2007 created [Manuel Guidon]
!> \author Manuel Guidon
! *****************************************************************************
  SUBROUTINE hfx_set_distr_forces(ptr_to_distr, x_data, error)
    TYPE(hfx_distribution), DIMENSION(:), &
      POINTER                                :: ptr_to_distr
    TYPE(hfx_type), POINTER                  :: x_data
    TYPE(cp_error_type), INTENT(inout)       :: error

    CHARACTER(LEN=*), PARAMETER :: routineN = 'hfx_set_distr_forces', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: stat
    LOGICAL                                  :: failure

    DEALLOCATE(x_data%distribution_forces,STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
   
    ALLOCATE(x_data%distribution_forces(SIZE(ptr_to_distr)),STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    x_data%distribution_forces = ptr_to_distr

  END SUBROUTINE hfx_set_distr_forces

! *****************************************************************************
!> \brief - resets the maximum memory usage for a HFX calculation subtracting
!>          all relevant buffers from the input MAX_MEM value and add 10% of
!>          safety margin
!> \param memory_parameter Memory information
!> \param subtr_size_mb size of buffers in MB
!> \param error variable to control error logging, stopping,...
!>        see module cp_error_handling
!> \par History
!>      02.2009 created [Manuel Guidon]
!> \author Manuel Guidon
! *****************************************************************************
  SUBROUTINE hfx_reset_memory_usage_counter(memory_parameter, subtr_size_mb, error)
    TYPE(hfx_memory_type)                    :: memory_parameter
    INTEGER(int_8), INTENT(IN)               :: subtr_size_mb
    TYPE(cp_error_type), INTENT(inout)       :: error

    CHARACTER(LEN=*), PARAMETER :: &
      routineN = 'hfx_reset_memory_usage_counter', &
      routineP = moduleN//':'//routineN

    INTEGER(int_8)                           :: max_memory

    max_memory = memory_parameter%max_memory
    max_memory = max_memory - subtr_size_mb
    IF( max_memory <= 0 ) THEN
      memory_parameter%do_all_on_the_fly = .TRUE.
      memory_parameter%max_compression_counter = 0
    ELSE
      memory_parameter%do_all_on_the_fly = .FALSE.
      memory_parameter%max_compression_counter = max_memory*1024_int_8*128_int_8
    END IF
  END SUBROUTINE hfx_reset_memory_usage_counter

! *****************************************************************************
!> \brief - This routine prints some information on HFX 
!> \param x_data contains all relevant data structures for hfx runs
!> \param hfx_section HFX input section
!> \param i_rep current replica ID
!> \param error variable to control error logging, stopping,...
!>        see module cp_error_handling
!> \par History
!>      03.2008 created [Manuel Guidon]
!> \author Manuel Guidon
! *****************************************************************************
  SUBROUTINE hfx_print_info(x_data, hfx_section, i_rep, error)
    TYPE(hfx_type), POINTER                  :: x_data
    TYPE(section_vals_type), POINTER         :: hfx_section
    INTEGER, INTENT(IN)                      :: i_rep
    TYPE(cp_error_type), INTENT(inout)       :: error

    CHARACTER(LEN=*), PARAMETER :: routineN = 'hfx_print_info', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: iw
    REAL(dp)                                 :: rc_ang
    TYPE(cp_logger_type), POINTER            :: logger

    NULLIFY(logger)
    logger => cp_error_get_logger(error)

    iw = cp_print_key_unit_nr(logger,hfx_section,"HF_INFO",&
         extension=".scfLog",error=error)
    IF (iw>0) THEN
      WRITE (UNIT=iw,FMT="(/,(T3,A,T61,I20))")&
            "HFX_INFO| Replica ID:     ",i_rep
      WRITE (UNIT=iw,FMT="((T3,A,T73,ES8.1))")&
            "HFX_INFO| EPS_SCHWARZ:     ", x_data%screening_parameter%eps_schwarz
      WRITE (UNIT=iw,FMT="((T3,A,T73,ES8.1))")&
            "HFX_INFO| EPS_SCHWARZ_FORCES     ", x_data%screening_parameter%eps_schwarz_forces
      WRITE (UNIT=iw,FMT="((T3,A,T73,ES8.1))")&
            "HFX_INFO| EPS_STORAGE_SCALING:     ", x_data%memory_parameter%eps_storage_scaling
      WRITE (UNIT=iw,FMT="((T3,A,T61,I20))")&
            "HFX_INFO| NBINS:     ", x_data%load_balance_parameter%nbins
      WRITE (UNIT=iw,FMT="((T3,A,T61,I20))")&
            "HFX_INFO| BLOCK_SIZE:     ", x_data%load_balance_parameter%block_size
      WRITE(iw,'(T3,A,T61,F20.10)') & 
            "HFX_INFO| FRACTION:     ", x_data%general_parameter%fraction
      SELECT CASE (x_data%potential_parameter%potential_type)
        CASE(do_hfx_potential_coulomb)
          WRITE (UNIT=iw,FMT="((T3,A,T74,A))")&
            "HFX_INFO| Interaction Potential:     ", "COULOMB"
        CASE(do_hfx_potential_short)
          WRITE (UNIT=iw,FMT="((T3,A,T71,A))")&
            "HFX_INFO| Interaction Potential:    ", "SHORTRANGE"
          WRITE(iw,'(T3,A,T61,F20.10)') &
            "HFX_INFO| Omega:     ", x_data%potential_parameter%omega
        CASE(do_hfx_potential_long)
          WRITE (UNIT=iw,FMT="((T3,A,T72,A))")&
            "HFX_INFO| Interaction Potential:     ", "LONGRANGE"
          WRITE(iw,'(T3,A,T61,F20.10)') &
            "HFX_INFO| Omega:     ", x_data%potential_parameter%omega
        CASE(do_hfx_potential_mix_cl)
          WRITE (UNIT=iw,FMT="((T3,A,T75,A))")&
            "HFX_INFO| Interaction Potential:     ", "MIX_CL"
          WRITE(iw,'(T3,A,T61,F20.10)') &
            "HFX_INFO| Omega:     ", x_data%potential_parameter%omega
          WRITE(iw,'(T3,A,T61,F20.10)') &
            "HFX_INFO| SCALE_COULOMB:     ", x_data%potential_parameter%scale_coulomb
          WRITE(iw,'(T3,A,T61,F20.10)') &
            "HFX_INFO| SCALE_LONGRANGE:     ", x_data%potential_parameter%scale_longrange
        CASE(do_hfx_potential_gaussian)
          WRITE (UNIT=iw,FMT="((T3,A,T73,A))")&
            "HFX_INFO| Interaction Potential:     ", "GAUSSIAN"
          WRITE(iw,'(T3,A,T61,F20.10)') &
            "HFX_INFO| Omega:     ", x_data%potential_parameter%omega
        CASE(do_hfx_potential_mix_lg)
          WRITE (UNIT=iw,FMT="((T3,A,T75,A))")&
            "HFX_INFO| Interaction Potential:    ", "MIX_LG"
          WRITE(iw,'(T3,A,T61,F20.10)') &
            "HFX_INFO| Omega:     ", x_data%potential_parameter%omega
          WRITE(iw,'(T3,A,T61,F20.10)') &
            "HFX_INFO| SCALE_LONGRANGE:     ", x_data%potential_parameter%scale_longrange
          WRITE(iw,'(T3,A,T61,F20.10)') &
            "HFX_INFO| SCALE_GAUSSIAN:    ", x_data%potential_parameter%scale_gaussian
        CASE(do_hfx_potential_id)
          WRITE (UNIT=iw,FMT="((T3,A,T73,A))")&
            "HFX_INFO| Interaction Potential:    ", "IDENTITY"
        CASE(do_hfx_potential_truncated)
          WRITE (UNIT=iw,FMT="((T3,A,T72,A))")&
            "HFX_INFO| Interaction Potential:    ", "TRUNCATED"
          rc_ang = cp_unit_from_cp2k(x_data%potential_parameter%cutoff_radius,"angstrom",error=error)
          WRITE(iw,'(T3,A,T61,F20.10)') &
            "HFX_INFO| Cutoff Radius [angstrom]:     ", rc_ang
      END SELECT
      IF( x_data%periodic_parameter%do_periodic) THEN
        IF( x_data%periodic_parameter%mode == -1 ) THEN
          WRITE (UNIT=iw,FMT="((T3,A,T77,A))")&
              "HFX_INFO| NUMBER_OF_SHELLS:     ", "AUTO"
        ELSE
          WRITE (UNIT=iw,FMT="((T3,A,T61,I20))")&
            "HFX_INFO| NUMBER_OF_SHELLS:     ", x_data%periodic_parameter%mode
        END IF
        WRITE (UNIT=iw,FMT="((T3,A,T61,I20))")&
            "HFX_INFO| Number of periodic shells considered:     ", x_data%periodic_parameter%number_of_shells
        WRITE (UNIT=iw,FMT="((T3,A,T61,I20)),/")&
              "HFX_INFO| Number of periodic cells considered:     ",SIZE(x_data%neighbor_cells)
      ELSE
        WRITE (UNIT=iw,FMT="((T3,A,T77,A))")&
            "HFX_INFO| Number of periodic shells considered:     ", "NONE"
        WRITE (UNIT=iw,FMT="((T3,A,T77,A)),/")&
              "HFX_INFO| Number of periodic cells considered:     ", "NONE"
      END IF
    END IF
    CALL cp_print_key_finished_output(iw,logger,hfx_section,&
                                     "HF_INFO", error=error)
  END SUBROUTINE hfx_print_info

END MODULE hfx_types

